QWtoASCII    proc

 comment    * ------------------------------------------------------------
            Convert positive QWORD to ASCII string (no sign) by IanB.

            On entry:
                EAX/EDX = value to convert Lo/Hi
                ECX = pointer to output string buffer (21 chars minimum)
            On exit:
                EAX = count of chars written
                ECX = pointer to null-termination byte in output string
                      in case we want to add more chars afterwards
                EDX = original pointer to string buffer
                string buffer contains the null-terminated ASCII string
            EBX/EBP/EDI/ESI preserved
            ------------------------------------------------------------ *

        push    ebx
        push    esi
        mov     esi, edx                    ; hold good values in EDI/ESI
        push    edi
        mov     edi, eax
        push    ecx                         ; save value
        mov     bl, "0"
        test    edx, edx                    ; can we skip the slow top DWORD stuff
        jz      TestLowDWORD

        mov     bh, bl                      ; keep for convenience
        cmp     esi, 2H
        jb      Digit09
        jz      Digit10a
        cmp     edx, 17H
        jb      Digit10
        jz      Digit11a
        cmp     esi, 0e8H
        jb      Digit11
        jz      Digit12a
        cmp     edx, 918H
        jb      Digit12
        jz      Digit13a
        cmp     esi, 5af3H
        jb      Digit13
        jz      Digit14a
        cmp     edx, 38d7eH
        jb      Digit14
        jz      Digit15a
        cmp     esi, 2386f2H
        jb      Digit15
        jz      Digit16a
        cmp     edx, 1634578H
        jb      Digit16
        jz      Digit17a
        cmp     esi, 0de0b6b3H
        jb      Digit17
        jz      Digit18a
        cmp     edx, 8ac72304H
        jb      Digit18
        ja      Digit19
        cmp     eax, 89e80000H
        jb      Digit18

Digit19:
        sub     eax, 89e80000H              ; 10,000,000,000,000,000,000
        sbb     edx, 8ac72304H
        add     bl, 1                       ; max value of 19th digit is 1
                                            ; so no need to loop
        mov     edi, eax                    ; copy good values
        mov     esi, edx
        mov     [ecx], bl
        mov     bl, bh
        add     ecx, 1
        jmp     Digit18

Digit18a:
        cmp     eax, 0a7640000H
        jb      Digit17

Digit18:
        sub     eax, 0a7640000H             ; 1,000,000,000,000,000,000
        sbb     edx, 0de0b6b3H
        jb      @F
        add     bl, 1
        mov     edi, eax                    ; save good values
        mov     esi, edx
        jmp     Digit18
@@:
        mov     eax, edi                    ; copy good values to test
        mov     edx, esi
        mov     [ecx], bl
        mov     bl, bh
        add     ecx, 1
        jmp     Digit17

Digit17a:
        cmp     eax, 5d8a0000H
        jb      Digit16

Digit17:
        sub     eax, 5d8a0000H              ; 100,000,000,000,000,000
        sbb     edx, 1634578H
        jb      @F
        add     bl, 1
        mov     edi, eax                    ; save good values
        mov     esi, edx
        jmp     Digit17
@@:
        mov     eax, edi                    ; copy good values to test
        mov     edx, esi
        mov     [ecx], bl
        mov     bl, bh
        add     ecx, 1
        jmp     Digit16

Digit16a:
        cmp     eax, 6fc10000H
        jb      Digit15

Digit16:
        sub     eax, 6fc10000H              ; 10,000,000,000,000,000
        sbb     edx, 2386f2H
        jb      @F
        add     bl, 1
        mov     edi, eax                    ; save good values
        mov     esi, edx
        jmp     Digit16
@@:
        mov     eax, edi                    ; copy good values to test
        mov     edx, esi
        mov     [ecx], bl
        mov     bl, bh
        add     ecx, 1
        jmp     Digit15

Digit15a:
        cmp     eax, 0a4c68000H
        jb      Digit14

Digit15:
        sub     eax, 0a4c68000H             ; 1,000,000,000,000,000
        sbb     edx, 38d7eH
        jb      @F
        add     bl, 1
        mov     edi, eax                    ; save good values
        mov     esi, edx
        jmp     Digit15
@@:
        mov     eax, edi                    ; copy good values to test
        mov     edx, esi
        mov     [ecx], bl
        mov     bl, bh
        add     ecx, 1
        jmp     Digit14

Digit14a:
        cmp     eax, 107a4000H
        jb      Digit13

Digit14:
        sub     eax, 107a4000H              ; 100,000,000,000,000
        sbb     edx, 5af3H
        jb      @F
        add     bl, 1
        mov     edi, eax                    ; save good values
        mov     esi, edx
        jmp     Digit14
@@:
        mov     eax, edi                    ; copy good values to test
        mov     edx, esi
        mov     [ecx], bl
        mov     bl, bh
        add     ecx, 1
        jmp     Digit13

Digit13a:
        cmp     eax, 4e72a000H
        jb      Digit12

Digit13:
        sub     eax, 4e72a000H              ; 10,000,000,000,000
        sbb     edx, 918H
        jb      @F
        add     bl, 1
        mov     edi, eax                    ; save good values
        mov     esi, edx
        jmp     Digit13
@@:
        mov     eax, edi                    ; copy good values to test
        mov     edx, esi
        mov     [ecx], bl
        mov     bl, bh
        add     ecx, 1
        jmp     Digit12

Digit12a:
        cmp     eax, 0d4a51000H
        jb      Digit11

Digit12:
        sub     eax, 0d4a51000H             ; 1,000,000,000,000
        sbb     edx, 0e8H
        jb      @F
        add     bl, 1
        mov     edi, eax                    ; save good values
        mov     esi, edx
        jmp     Digit12
@@:
        mov     eax, edi                    ; copy good values to test
        mov     edx, esi
        mov     [ecx], bl
        mov     bl, bh
        add     ecx, 1
        jmp     Digit11

Digit11a:
        cmp     eax, 4876e800H
        jb      Digit10

Digit11:
        sub     eax, 4876e800H              ; 100,000,000,000
        sbb     edx, 17H
        jb      @F
        add     bl, 1
        mov     edi, eax                    ; save good values
        mov     esi, edx
        jmp     Digit11
@@:
        mov     eax, edi                    ; copy good values to test
        mov     edx, esi
        mov     [ecx], bl
        mov     bl, bh
        add     ecx, 1
        jmp     Digit10

Digit10a:
        cmp     eax, 540be400H
        jb      Digit09

Digit10:
        sub     eax, 540be400H              ; 10,000,000,000
        sbb     edx, 2H
        jb      @F
        add     bl, 1
        mov     edi, eax                    ; save good values
        mov     esi, edx
        jmp     Digit10
@@:
        mov     eax, edi                    ; copy good values to test
        mov     [ecx], bl
        mov     bl, bh
        add     ecx, 1

Digit09:
        mov     edx, 1                      ; convenience number
@@:
        sub     eax, 1000000000
        sbb     esi, 0
        jb      @F
        add     bl, dl
        mov     edi, eax                    ; save good value
        jmp     @B

TestLowDWORD:
        mov     edx, 1                      ; convenience number
        cmp     eax, 10
        jb      Digit00
        cmp     edi, 100
        jb      Digit01
        cmp     eax, 1000
        jb      Digit02
        cmp     edi, 10000
        jb      Digit03
        cmp     eax, 100000
        jb      Digit04
        cmp     edi, 1000000
        jb      Digit05
        cmp     eax, 10000000
        jb      Digit06
        cmp     edi, 100000000
        jb      Digit07
        sub     eax, 1000000000
        jb      Digit08

Digit09a:
        mov     edi, eax                    ; save good value, no high DWORD
        add     bl, dl
        sub     eax, 1000000000             ; only 3 iterations of this loop max
        jae     Digit09a
@@:
        mov     [ecx], bl
        add     ecx, edx

Digit08:
        mov     eax, (-900000000)
        mov     esi, edi                    ; copy good value to test
        mov     bl, "0"
        mov     bh, "9"
        add     eax, edi                    ; check if digit = 9
        jns     FromTop08
@@:
        sub     esi, 100000000
        jb      @F
        add     bl, dl
        sub     bh, dl
        add     eax, 100000000
        mov     edi, esi                    ; save good value
        js      @B

FromTop08:
        mov     bl, bh
        mov     edi, eax
@@:
        mov     [ecx], bl
        add     ecx, edx

Digit07:
        mov     eax, (-90000000)
        mov     esi, edi                    ; copy good value to test
        mov     bl, "0"
        mov     bh, "9"
        add     eax, edi                    ; check if digit = 9
        jns     FromTop07
@@:
        sub     esi, 10000000
        jb      @F
        add     bl, dl
        sub     bh, dl
        add     eax, 10000000
        mov     edi, esi                    ; save good value
        js      @B

FromTop07:
        mov     bl, bh
        mov     edi, eax
@@:
        mov     [ecx], bl
        add     ecx, edx

Digit06:
        mov     eax, (-9000000)
        mov     esi, edi                    ; copy good value to test
        mov     bl, "0"
        mov     bh, "9"
        add     eax, edi                    ; check if digit = 9
        jns     FromTop06
@@:
        sub     esi, 1000000
        jb      @F
        add     bl, dl
        sub     bh, dl
        add     eax, 1000000
        mov     edi, esi                    ; save good value
        js      @B

FromTop06:
        mov     bl, bh
        mov     edi, eax
@@:
        mov     [ecx], bl
        add     ecx, edx

Digit05:
        mov     eax, (-900000)
        mov     esi, edi                    ; copy good value to test
        mov     bl, "0"
        mov     bh, "9"
        add     eax, edi                    ; check if digit = 9
        jns     FromTop05
@@:
        sub     esi, 100000
        jb      @F
        add     bl, dl
        sub     bh, dl
        add     eax, 100000
        mov     edi, esi                    ; save good value
        js      @B

FromTop05:
        mov     bl, bh
        mov     edi, eax
@@:
        mov     [ecx], bl
        add     ecx, edx

Digit04:
        mov     eax, (-90000)
        mov     esi, edi                    ; copy good value to test
        mov     bl, "0"
        mov     bh, "9"
        add     eax, edi                    ; check if digit = 9
        jns     FromTop04
@@:
        sub     esi, 10000
        jb      @F
        add     bl, dl
        sub     bh, dl
        add     eax, 10000
        mov     edi, esi                    ; save good value
        js      @B

FromTop04:
        mov     bl, bh
        mov     edi, eax
@@:
        mov     [ecx], bl
        add     ecx, edx

Digit03:
        mov     eax, (-9000)
        mov     esi, edi                    ; copy good value to test
        mov     bl, "0"
        mov     bh, "9"
        add     eax, edi                    ; check if digit = 9
        jns     FromTop03
@@:
        sub     esi, 1000
        jb      @F
        add     bl, dl
        sub     bh, dl
        add     eax, 1000
        mov     edi, esi                    ; save good value
        js      @B

FromTop03:
        mov     bl, bh
        mov     edi, eax
@@:
        mov     [ecx], bl
        add     ecx, edx

Digit02:
        mov     eax, (-900)
        mov     esi, edi                    ; copy good value to test
        mov     bl, "0"
        mov     bh, "9"
        add     eax, edi                    ; check if digit = 9
        jns     FromTop02
@@:
        sub     esi, 100
        jb      @F
        add     bl, dl
        sub     bh, dl
        add     eax, 100
        mov     edi, esi                    ; save good value
        js      @B

FromTop02:
        mov     bl, bh
        mov     edi, eax
@@:
        mov     [ecx], bl
        add     ecx, edx

Digit01:
        mov     esi, (-90)
        mov     bl, "0"
        mov     bh, "9"
        add     esi, edi                    ; check if digit = 9
        jns     FromTop01
@@:
        mov     eax, edi                    ; save good value
        sub     edi, 10
        jb      @F
        add     bl, dl
        sub     bh, dl
        add     esi, 10
        js      @B

FromTop01:
        mov     bl, bh
        mov     eax, esi                    ; copy good final digit
@@:
        mov     [ecx], bl
        add     ecx, edx

Digit00:
        add     al, "0"
        mov     ebx, edx                    ; EBX = 1
        pop     edx                         ; retrieve original string pointer
        mov     [ecx], ax                   ; add null-termination
        add     ecx, ebx                    ; point to null-termination byte
        pop     edi
        mov     eax, ecx
        pop     esi
        sub     eax, edx                    ; get count of chars written
        pop     ebx
        ret

QWtoASCII endp
